__all__ = ['ANTsTransform',
           'set_ants_transform_parameters',
           'get_ants_transform_parameters',
           'get_ants_transform_fixed_parameters',
           'set_ants_transform_fixed_parameters',
           'apply_ants_transform',
           'apply_ants_transform_to_point',
           'apply_ants_transform_to_vector',
           'apply_ants_transform_to_image',
           'invert_ants_transform',
           'compose_ants_transforms',
           'transform_index_to_physical_point',
           'transform_physical_point_to_index']

import numpy as np

import ants
from ants.internal import get_lib_fn, short_ptype


class ANTsTransform(object):

    def __init__(self, precision='float', dimension=3, transform_type='AffineTransform', pointer=None):
        """
        NOTE: This class should never be initialized directly by the user.

        Initialize an ANTsTransform object.

        Arguments
        ---------
        pixeltype : string

        dimension : integer

        transform_type : stirng

        pointer : py::capsule (optional)
        """
        self.precision = precision
        self.dimension = dimension
        self.transform_type = transform_type
        self.type = transform_type

        if pointer is None:
            libfn = get_lib_fn('newAntsTransform%s%i' % (short_ptype(precision), dimension))
            pointer = libfn(precision, dimension, transform_type)

        self.pointer = pointer

        # suffix to append to all c++ library function calls
        self._libsuffix = '%s%i' % (short_ptype(self.precision), self.dimension)

    @property
    def parameters(self):
        """ Get parameters of transform """
        libfn = get_lib_fn('getTransformParameters')
        return np.asarray(libfn(self.pointer), order='F')#.reshape((self.dimension, self.dimension+1), order='F')

    def set_parameters(self, parameters):
        """ Set parameters of transform """
        if not isinstance(parameters, np.ndarray):
            parameters = np.asarray(parameters)

        # if in two dimensions, flatten with fortran ordering
        if parameters.ndim > 1:
            parameters = parameters.flatten(order='F')

        libfn = get_lib_fn('setTransformParameters')
        libfn(self.pointer, parameters.tolist())

    @property
    def fixed_parameters(self):
        """ Get parameters of transform """
        libfn = get_lib_fn('getTransformFixedParameters')
        return np.asarray(libfn(self.pointer))

    def set_fixed_parameters(self, parameters):
        """ Set parameters of transform """
        if not isinstance(parameters, np.ndarray):
            parameters = np.asarray(parameters)

        libfn = get_lib_fn('setTransformFixedParameters')
        libfn(self.pointer, parameters.tolist())

    def invert(self):
        """ Invert the transform """
        libfn = get_lib_fn('inverseTransform')
        inv_tx_ptr = libfn(self.pointer)

        new_tx = ANTsTransform(precision=self.precision, dimension=self.dimension,
                                transform_type=self.transform_type, pointer=inv_tx_ptr)

        return new_tx

    def apply(self, data, data_type='point', reference=None, **kwargs):
        """
        Apply transform to data

        """
        if data_type == 'point':
            return self.apply_to_point(data)
        elif data_type == 'vector':
            return self.apply_to_vector(data)
        elif data_type == 'image':
            return self.apply_to_image(data, reference, **kwargs)

    def apply_to_point(self, point):
        """
        Apply transform to a point

        Arguments
        ---------
        point : list/tuple
            point to which the transform will be applied

        Returns
        -------
        list : transformed point

        Example
        -------
        >>> import ants
        >>> tx = ants.new_ants_transform()
        >>> params = tx.parameters
        >>> tx.set_parameters(params*2)
        >>> pt2 = tx.apply_to_point((1,2,3)) # should be (2,4,6)
        """
        libfn = get_lib_fn('transformPoint')
        return tuple(libfn(self.pointer, point))

    def apply_to_vector(self, vector):
        """
        Apply transform to a vector

        Arguments
        ---------
        vector : list/tuple
            vector to which the transform will be applied

        Returns
        -------
        list : transformed vector
        """
        if isinstance(vector, np.ndarray):
            vector = vector.tolist()

        libfn = get_lib_fn('transformVector')
        return np.asarray(libfn(self.pointer, vector))

    def apply_to_image(self, image, reference=None, interpolation='linear'):
        """
        Apply transform to an image

        Arguments
        ---------
        image : ANTsImage
            image to which the transform will be applied

        reference : ANTsImage
            target space for transforming image

        interpolation : string
            type of interpolation to use. Options are:
                linear
                nearestneighbor
                multilabel
                gaussian
                bspline
                cosinewindowedsinc
                welchwindowedsinc
                hammingwindoweddinc
                lanczoswindowedsinc
                genericlabel

        Returns
        -------
        list : transformed vector

        """
        if reference is None:
            reference = image.clone()
        
        interpolation = interpolation.lower()

        tform_fn = get_lib_fn('transformImage')
        reference = reference.clone(image.pixeltype)

        img_ptr = tform_fn(self.pointer, image.pointer, reference.pointer, interpolation)
        return ants.from_pointer(img_ptr)

    def __repr__(self):
        s = "ANTsTransform\n" +\
            '\t {:<10} : {}\n'.format('Type', self.type)+\
            '\t {:<10} : {}\n'.format('Dimension', self.dimension)+\
            '\t {:<10} : {}\n'.format('Precision', self.precision)
        return s

# verbose functions for ANTsR compatibility
def set_ants_transform_parameters(transform, parameters):
    """
    Set parameters of an ANTsTransform

    ANTsR function: `setAntsrTransformParameters`
    """
    transform.set_parameters(parameters)

def get_ants_transform_parameters(transform):
    """
    Get parameters of an ANTsTransform

    ANTsR function: `getAntsrTransformParameters`
    """
    return transform.parameters

def set_ants_transform_fixed_parameters(transform, parameters):
    """
    Set fixed parameters of an ANTsTransform

    ANTsR function: `setAntsrTransformFixedParameters`
    """
    transform.set_fixed_parameters(parameters)

def get_ants_transform_fixed_parameters(transform):
    """
    Get fixed parameters of an ANTsTransform

    ANTsR function: `getAntsrTransformFixedParameters`
    """
    return transform.fixed_parameters


def apply_ants_transform(transform, data, data_type="point", reference=None, **kwargs):
    """
    Apply ANTsTransform to data

    ANTsR function: `applyAntsrTransform`

    Arguments
    ---------
    transform : ANTsTransform
        transform to apply to image

    data : ndarray/list/tuple
        data to which transform will be applied

    data_type : string
        type of data
        Options :
            'point'
            'vector'
            'image'

    reference : ANTsImage
        target space for transforming image

    kwargs : kwargs
        additional options passed to `apply_ants_transform_to_image`

    Returns
    -------
    ANTsImage if data_type == 'image'
    OR
    tuple if data_type == 'point' or data_type == 'vector'
    """
    return transform.apply(data, data_type, reference, **kwargs)


def apply_ants_transform_to_point(transform, point):
    """
    Apply transform to a point

    ANTsR function: `applyAntsrTransformToPoint`

    Arguments
    ---------
    point : list/tuple
        point to which the transform will be applied

    Returns
    -------
    tuple : transformed point

    Example
    -------
    >>> import ants
    >>> tx = ants.new_ants_transform()
    >>> params = tx.parameters
    >>> tx.set_parameters(params*2)
    >>> pt2 = tx.apply_to_point((1,2,3)) # should be (2,4,6)
    """
    return transform.apply_to_point(point)


def apply_ants_transform_to_vector(transform, vector):
    """
    Apply transform to a vector

    ANTsR function: `applyAntsrTransformToVector`

    Arguments
    ---------
    vector : list/tuple
        vector to which the transform will be applied

    Returns
    -------
    tuple : transformed vector
    """
    return transform.apply_to_vector(vector)


def apply_ants_transform_to_image(transform, image, reference, interpolation='linear'):
    """
    Apply transform to an image

    ANTsR function: `applyAntsrTransformToImage`

    Arguments
    ---------
    image : ANTsImage
        image to which the transform will be applied

    reference : ANTsImage
        reference image

    interpolation : string
        type of interpolation to use. Options are:
        linear
        nearestneighbor
        multilabel
        gaussian
        bspline
        cosinewindowedsinc
        welchwindowedsinc
        hammingwindoweddinc
        lanczoswindowedsinc
        genericlabel

    Returns
    -------
    list : transformed vector

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data("r16")).clone('float')
    >>> tx = ants.new_ants_transform(dimension=2)
    >>> tx.set_parameters((0.9,0,0,1.1,10,11))
    >>> img2 = tx.apply_to_image(img, img)
    """
    return transform.apply_to_image(image, reference, interpolation)


def invert_ants_transform(transform):
    """
    Invert ANTsTransform

    ANTsR function: `invertAntsrTransform`

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data("r16")).clone('float')
    >>> tx = ants.new_ants_transform(dimension=2)
    >>> tx.set_parameters((0.9,0,0,1.1,10,11))
    >>> img_transformed = tx.apply_to_image(img, img)
    >>> inv_tx = tx.invert()
    >>> img_orig = inv_tx.apply_to_image(img_transformed, img_transformed)
    """
    return transform.invert()

def compose_ants_transforms(transform_list):
    """
    Compose multiple ANTsTransform's together

    ANTsR function: `composeAntsrTransforms`

    Arguments
    ---------
    transform_list : list/tuple of ANTsTransform object
        list of transforms to compose together

    Returns
    -------
    ANTsTransform
        one transform that contains all given transforms

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data("r16")).clone('float')
    >>> tx = ants.new_ants_transform(dimension=2)
    >>> tx.set_parameters((0.9,0,0,1.1,10,11))
    >>> inv_tx = tx.invert()
    >>> single_tx = ants.compose_ants_transforms([tx, inv_tx])
    >>> img_orig = single_tx.apply_to_image(img, img)
    >>> rRotGenerator = ants.contrib.RandomRotate2D( ( 0, 40 ), reference=img )
    >>> rShearGenerator=ants.contrib.RandomShear2D( (0,50), reference=img )
    >>> tx1 = rRotGenerator.transform()
    >>> tx2 = rShearGenerator.transform()
    >>> rSrR = ants.compose_ants_transforms([tx1, tx2])
    >>> rSrR.apply_to_image( img )
    """
    precision = transform_list[0].precision
    dimension = transform_list[0].dimension

    for tx in transform_list:
        if precision != tx.precision:
            raise ValueError('All transforms must have the same precision')
        if dimension != tx.dimension:
            raise ValueError('All transforms must have the same dimension')

    tx_ptr_list = list(reversed([tf.pointer for tf in transform_list]))
    libfn = get_lib_fn('composeTransforms%s' % (transform_list[0]._libsuffix))
    itk_composed_tx = libfn(tx_ptr_list, precision, dimension)
    return ANTsTransform(precision=precision, dimension=dimension,
                        transform_type='CompositeTransform', pointer=itk_composed_tx)


def transform_index_to_physical_point(image, index):
    """
    Get spatial point from index of an image.

    ANTsR function: `antsTransformIndexToPhysicalPoint`

    Arguments
    ---------
    img : ANTsImage
        image to get values from

    index : list or tuple or numpy.ndarray
        location in image

    Returns
    -------
    tuple

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> img = ants.make_image((10,10),np.random.randn(100))
    >>> pt = ants.transform_index_to_physical_point(img, (2,2))
    """
    if not ants.is_image(image):
        raise ValueError('image must be ANTsImage type')

    if isinstance(index, np.ndarray):
        index = index.tolist()

    if not isinstance(index, (tuple,list)):
        raise ValueError('index must be tuple or list')

    if len(index) != image.dimension:
        raise ValueError('len(index) != image.dimension')

    index = [i+1 for i in index]
    ndim = image.dimension
    ptype = image.pixeltype
    libfn = get_lib_fn('TransformIndexToPhysicalPoint')
    point = libfn(image.pointer, [list(index)])

    return np.array(point[0])


def transform_physical_point_to_index(image, point):
    """
    Get index from spatial point of an image.

    ANTsR function: `antsTransformPhysicalPointToIndex`

    Arguments
    ---------
    image : ANTsImage
        image to get values from

    point : list or tuple or numpy.ndarray
        point in image

    Returns
    -------
    tuple

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> img = ants.make_image((10,10),np.random.randn(100))
    >>> idx = ants.transform_physical_point_to_index(img, (2,2))
    >>> img.set_spacing((2,2))
    >>> idx2 = ants.transform_physical_point_to_index(img, (4,4))
    """
    if not ants.is_image(image):
        raise ValueError('image must be ANTsImage type')

    if isinstance(point, np.ndarray):
        point = point.tolist()

    if not isinstance(point, (list,tuple)):
        raise ValueError('point must be list, tuple, or np.ndarray')

    if len(point) != image.dimension:
        raise ValueError('len(index) != image.dimension')

    ndim = image.dimension
    ptype = image.pixeltype
    libfn = get_lib_fn('TransformPhysicalPointToIndex')
    index = libfn(image.pointer, [list(point)])
    index = [i-1 for i in index[0]]

    return np.array(index)



__all__ = ["add_noise_to_image"]


import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def add_noise_to_image(image,
                       noise_model,
                       noise_parameters
):
    """
    Add noise to an image using additive Gaussian, salt-and-pepper,
    shot, or speckle noise.

    Arguments
    ---------
    image : ANTsImage
        scalar image.

    noise_model : string
        'additivegaussian', 'saltandpepper', 'shot', or 'speckle'.

    noise_parameters : tuple or array or float
        'additivegaussian': (mean, standardDeviation)
        'saltandpepper': (probability, saltValue, pepperValue)
        'shot': scale
        'speckle': standardDeviation

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> noise_image = ants.add_noise_to_image(image, 'additivegaussian', (0.0, 1.0))
    >>> noise_image = ants.add_noise_to_image(image, 'saltandpepper', (0.1, 0.0, 100.0))
    >>> noise_image = ants.add_noise_to_image(image, 'shot', 1.0)
    >>> noise_image = ants.add_noise_to_image(image, 'speckle', 1.0)
    """

    image_dimension = image.dimension

    if noise_model == 'additivegaussian':
        if len(noise_parameters) != 2:
            raise ValueError("Incorrect number of parameters.")

        libfn = get_lib_fn("additiveGaussianNoiseF%i" % image_dimension)
        noise = libfn(image.pointer, noise_parameters[0], noise_parameters[1])
        output_image = ants.from_pointer(noise).clone('float')
        return output_image
    elif noise_model == 'saltandpepper':
        if len(noise_parameters) != 3:
            raise ValueError("Incorrect number of parameters.")

        libfn = get_lib_fn("saltAndPepperNoiseF%i" % image_dimension)
        noise = libfn(image.pointer, noise_parameters[0], noise_parameters[1], noise_parameters[2])
        output_image = ants.from_pointer(noise).clone('float')
        return output_image
    elif noise_model == 'shot':
        if not isinstance(noise_parameters, (int, float)):
            raise ValueError("Incorrect parameter specification.")

        libfn = get_lib_fn("shotNoiseF%i" % image_dimension)
        noise = libfn(image.pointer, noise_parameters)
        output_image = ants.from_pointer(noise).clone('float')
        return output_image
    elif noise_model == 'speckle':
        if not isinstance(noise_parameters, (int, float)):
            raise ValueError("Incorrect parameter specification.")

        libfn = get_lib_fn("speckleNoiseF%i" % image_dimension)
        noise = libfn(image.pointer, noise_parameters)
        output_image = ants.from_pointer(noise).clone('float')
        return output_image
    else:
        raise ValueError("Unknown noise model.")    



import ants
from ants import lib

short_ptype_map = {"unsigned char": "UC",
                    "unsigned int": "UI",
                    "float": "F",
                    "double": "D"}

def infer_dtype(dtype):
    # supported dtypes: uint8, uint32, float32, float64
    exchange_map = {
        'bool': 'uint8',
        'int8': 'uint32',
        'int16': 'uint32',
        'int32': 'uint32',
        'int64': 'uint32',
        'uint16': 'uint32',
        'uint64': 'uint32',
        'float16': 'float32'
    }
    new_dtype = exchange_map.get(str(dtype), dtype)
    return new_dtype

def short_ptype(pixeltype):
    return short_ptype_map[pixeltype]

def get_lib_fn(string):
    return getattr(lib, string)

def get_pointer_string(image):
    return lib.ptrstr(image.pointer)

def process_arguments(args):
    p_args = []
    if isinstance(args, dict):
        for argname, argval in args.items():
            
            # -MULTINAME- is used to support multiple items with the same key
            if "-MULTINAME-" in argname:
                argname = argname[: argname.find("-MULTINAME-")]

            if argval is not None:
                if len(argname) > 1:
                    p_args.append("--%s" % argname)
                else:
                    p_args.append("-%s" % argname)

                if ants.is_image(argval):
                    p_args.append(get_pointer_string(argval))
                elif isinstance(argval, list):
                    for av in argval:
                        if ants.is_image(av):
                            av = get_pointer_string(av)
                        elif str(arg) == "True":
                            av = str(1)
                        elif str(arg) == "False":
                            av = str(0)
                        p_args.append(av)
                else:
                    p_args.append(str(argval))

    elif isinstance(args, list):
        for arg in args:
            if ants.is_image(arg):
                pointer_string = get_pointer_string(arg)
                p_arg = pointer_string
            elif arg is None:
                pass
            elif str(arg) == "True":
                p_arg = str(1)
            elif str(arg) == "False":
                p_arg = str(0)
            else:
                p_arg = str(arg)
            p_args.append(p_arg)
    return p_args



"""
Functions for plotting ants images
"""


__all__ = [
    "plot_ortho"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


import numpy as np
import ants
from ants.decorators import image_method

@image_method
def plot_ortho(
    image,
    overlay=None,
    reorient=True,
    blend=False,
    # xyz arguments
    xyz=None,
    xyz_lines=True,
    xyz_color="red",
    xyz_alpha=0.6,
    xyz_linewidth=2,
    xyz_pad=5,
    orient_labels=True,
    # base image arguments
    alpha=1,
    cmap="Greys_r",
    # overlay arguments
    overlay_cmap="jet",
    overlay_alpha=0.9,
    cbar=False,
    cbar_length=0.8,
    cbar_dx=0.0,
    cbar_vertical=True,
    # background arguments
    black_bg=True,
    bg_thresh_quant=0.01,
    bg_val_quant=0.99,
    # scale/crop/domain arguments
    crop=False,
    scale=False,
    domain_image_map=None,
    # title arguments
    title=None,
    titlefontsize=24,
    title_dx=0,
    title_dy=0,
    # 4th panel text arguemnts
    text=None,
    textfontsize=24,
    textfontcolor="white",
    text_dx=0,
    text_dy=0,
    # save & size arguments
    filename=None,
    dpi=500,
    figsize=1.0,
    flat=False,
    transparent=True,
    resample=False,
    allow_xyz_change=True,
):
    """
    Plot an orthographic view of a 3D image

    Use mask_image and/or threshold_image to preprocess images to be be
    overlaid and display the overlays in a given range. See the wiki examples.

    ANTsR function: N/A

    Arguments
    ---------
    image : ANTsImage
        image to plot

    overlay : ANTsImage
        image to overlay on base image

    xyz : list or tuple of 3 integers
        selects index location on which to center display
        if given, solid lines will be drawn to converge at this coordinate.
        This is useful for pinpointing a specific location in the image.

    flat : boolean
        if true, the ortho image will be plot in one row
        if false, the ortho image will be a 2x2 grid with the bottom
            left corner blank

    cmap : string
        colormap to use for base image. See matplotlib.

    overlay_cmap : string
        colormap to use for overlay images, if applicable. See matplotlib.

    overlay_alpha : float
        level of transparency for any overlays. Smaller value means
        the overlay is more transparent. See matplotlib.

    cbar: boolean
        if true, a colorbar will be added to the plot

    cbar_length: float
        length of the colorbar relative to the image

    cbar_dx: float
        horizontal shift of the colorbar relative to the image

    cbar_vertical: boolean
        if true, the colorbar will be vertical, if false, it will be
        horizontal underneath the image

    axis : integer
        which axis to plot along if image is 3D

    black_bg : boolean
        if True, the background of the image(s) will be black.
        if False, the background of the image(s) will be determined by the
            values `bg_thresh_quant` and `bg_val_quant`.

    bg_thresh_quant : float
        if white_bg=True, the background will be determined by thresholding
        the image at the `bg_thresh` quantile value and setting the background
        intensity to the `bg_val` quantile value.
        This value should be in [0, 1] - somewhere around 0.01 is recommended.
            - equal to 1 will threshold the entire image
            - equal to 0 will threshold none of the image

    bg_val_quant : float
        if white_bg=True, the background will be determined by thresholding
        the image at the `bg_thresh` quantile value and setting the background
        intensity to the `bg_val` quantile value.
        This value should be in [0, 1]
            - equal to 1 is pure white
            - equal to 0 is pure black
            - somewhere in between is gray

    domain_image_map : ANTsImage
        this input ANTsImage or list of ANTsImage types contains a reference image
        `domain_image` and optional reference mapping named `domainMap`.
        If supplied, the image(s) to be plotted will be mapped to the domain
        image space before plotting - useful for non-standard image orientations.

    crop : boolean
        if true, the image(s) will be cropped to their bounding boxes, resulting
        in a potentially smaller image size.
        if false, the image(s) will not be cropped

    scale : boolean or 2-tuple
        if true, nothing will happen to intensities of image(s) and overlay(s)
        if false, dynamic range will be maximized when visualizing overlays
        if 2-tuple, the image will be dynamically scaled between these quantiles

    title : string
        add a title to the plot

    filename : string
        if given, the resulting image will be saved to this file

    dpi : integer
        determines resolution of image if saved to file. Higher values
        result in higher resolution images, but at a cost of having a
        larger file size

    resample : resample image in case of unbalanced spacing

    allow_xyz_change : boolean will attempt to adjust xyz after padding

    Example
    -------
    >>> import ants
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> ants.plot_ortho(mni, xyz=(100,100,100))
    >>> mni2 = mni.threshold_image(7000, mni.max())
    >>> ants.plot_ortho(mni, overlay=mni2)
    >>> ants.plot_ortho(mni, overlay=mni2, flat=True)
    >>> ants.plot_ortho(mni, overlay=mni2, xyz=(110,110,110), xyz_lines=False,
                        text='Lines Turned Off', textfontsize=22)
    >>> ants.plot_ortho(mni, mni2, xyz=(120,100,100),
                        text=' Example \nOrtho Text', textfontsize=26,
                        title='Example Ortho Title', titlefontsize=26)
    """

    def mirror_matrix(x):
        return x[::-1, :]

    def rotate270_matrix(x):
        return mirror_matrix(x.T)

    def reorient_slice(x, axis):
        return rotate270_matrix(x)

    # need this hack because of a weird NaN warning from matplotlib with overlays
    warnings.simplefilter("ignore")

    # handle `image` argument
    if isinstance(image, str):
        image = ants.image_read(image)
    if not ants.is_image(image):
        raise ValueError("image argument must be an ANTsImage")
    if image.dimension != 3:
        raise ValueError("Input image must have 3 dimensions!")

    # handle `overlay` argument
    if overlay is not None:
        if isinstance(overlay, str):
            overlay = ants.image_read(overlay)
        vminol = overlay.min()
        vmaxol = overlay.max()
        if not ants.is_image(overlay):
            raise ValueError("overlay argument must be an ANTsImage")
        if overlay.components > 1:
            raise ValueError("overlay cannot have more than one voxel component")
        if overlay.dimension != 3:
            raise ValueError("Overlay image must have 3 dimensions!")

        if not ants.image_physical_space_consistency(image, overlay):
            overlay = ants.resample_image_to_target(overlay, image, interp_type="linear")

    if blend:
        if alpha == 1:
            alpha = 0.5
        image = image * alpha + overlay * (1 - alpha)
        overlay = None
        alpha = 1.0

    if image.pixeltype not in {"float", "double"}:
        scale = False  # turn off scaling if image is discrete

    # reorient images
    if reorient != False:
        if reorient == True:
            reorient = "RPI"
        image = image.reorient_image2("RPI")
        if overlay is not None:
            overlay = overlay.reorient_image2("RPI")

    # handle `slices` argument
    if xyz is None:
        xyz = [int(s / 2) for s in image.shape]
    for i in range(3):
        if xyz[i] is None:
            xyz[i] = int(image.shape[i] / 2)

    # resample image if spacing is very unbalanced
    spacing = [s for i, s in enumerate(image.spacing)]
    if (max(spacing) / min(spacing)) > 3.0 and resample:
        new_spacing = (1, 1, 1)
        image = image.resample_image(tuple(new_spacing))
        if overlay is not None:
            overlay = overlay.resample_image(tuple(new_spacing))
        xyz = [
            int(sl * (sold / snew)) for sl, sold, snew in zip(xyz, spacing, new_spacing)
        ]


    # potentially crop image
    if crop:
        plotmask = image.get_mask(cleanup=0)
        if plotmask.max() == 0:
            plotmask += 1
        image = image.crop_image(plotmask)
        if overlay is not None:
            overlay = overlay.crop_image(plotmask)

    # pad images
    if True:
        image, lowpad, uppad = image.pad_image(return_padvals=True)
        if allow_xyz_change:
            xyz = [v + l for v, l in zip(xyz, lowpad)]
        if overlay is not None:
            overlay = overlay.pad_image()


    # handle `domain_image_map` argument
    if domain_image_map is not None:
        if ants.is_image(domain_image_map):
            tx = ants.new_ants_transform(
                precision="float",
                transform_type="AffineTransform",
                dimension=image.dimension,
            )
            image = ants.apply_ants_transform_to_image(tx, image, domain_image_map)
            if overlay is not None:
                overlay = ants.apply_ants_transform_to_image(
                    tx, overlay, domain_image_map, interpolation="linear"
                )
        else:
            raise Exception('The domain_image_map must be an image.')

    ## single-channel images ##
    if image.components == 1:

        # potentially find dynamic range
        if scale == True:
            vmin, vmax = image.quantile((0.05, 0.95))
        elif isinstance(scale, (list, tuple)):
            if len(scale) != 2:
                raise ValueError(
                    "scale argument must be boolean or list/tuple with two values"
                )
            vmin, vmax = image.quantile(scale)
        else:
            vmin = None
            vmax = None

        if not flat:
            nrow = 2
            ncol = 2
        else:
            nrow = 1
            ncol = 3

        fig = plt.figure(figsize=(9 * figsize, 9 * figsize))
        if title is not None:
            basey = 0.88 if not flat else 0.66
            basex = 0.5
            fig.suptitle(
                title, fontsize=titlefontsize, color=textfontcolor, x=basex + title_dx, y=basey + title_dy
            )

        gs = gridspec.GridSpec(
            nrow,
            ncol,
            wspace=0.0,
            hspace=0.0,
            top=1.0 - 0.5 / (nrow + 1),
            bottom=0.5 / (nrow + 1),
            left=0.5 / (ncol + 1),
            right=1 - 0.5 / (ncol + 1),
        )

        # pad image to have isotropic array dimensions
        imageReturn = image.clone()
        image = image.numpy()
        overlayReturn = None
        if overlay is not None:
            overlayReturn = overlay.clone()
            overlay = overlay.numpy()
            if overlay.dtype not in ["uint8", "uint32"]:
                overlay = np.ma.masked_where( np.abs(overlay) <= 1e-16, overlay)
#                overlay[np.abs(overlay) == 0] = np.nan

        yz_slice = reorient_slice(image[xyz[0], :, :], 0)
        ax = plt.subplot(gs[0, 0])
        ax.imshow(yz_slice, cmap=cmap, vmin=vmin, vmax=vmax)
        if overlay is not None:
            yz_overlay = reorient_slice(overlay[xyz[0], :, :], 0)
            ax.imshow(yz_overlay, alpha=overlay_alpha, cmap=overlay_cmap, vmin=vminol, vmax=vmaxol )
        if xyz_lines:
            # add lines
            l = mlines.Line2D(
                [xyz[1], xyz[1]],
                [xyz_pad, yz_slice.shape[0] - xyz_pad],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
            l = mlines.Line2D(
                [xyz_pad, yz_slice.shape[1] - xyz_pad],
                [yz_slice.shape[1] - xyz[2], yz_slice.shape[1] - xyz[2]],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
        if orient_labels:
            ax.text(
                0.5,
                0.98,
                "S",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                0.02,
                "I",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.98,
                0.5,
                "A",
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.02,
                0.5,
                "P",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
        ax.axis("off")

        xz_slice = reorient_slice(image[:, xyz[1], :], 1)
        ax = plt.subplot(gs[0, 1])
        ax.imshow(xz_slice, cmap=cmap, vmin=vmin, vmax=vmax)
        if overlay is not None:
            xz_overlay = reorient_slice(overlay[:, xyz[1], :], 1)
            ax.imshow(xz_overlay, alpha=overlay_alpha, cmap=overlay_cmap, vmin=vminol, vmax=vmaxol )

        if xyz_lines:
            # add lines
            l = mlines.Line2D(
                [xz_slice.shape[0] - xyz[0], xz_slice.shape[0] - xyz[0]],
                [xyz_pad, xz_slice.shape[0] - xyz_pad],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
            l = mlines.Line2D(
                [xyz_pad, xz_slice.shape[1] - xyz_pad],
                [xz_slice.shape[1] - xyz[2], xz_slice.shape[1] - xyz[2]],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
        if orient_labels:
            ax.text(
                0.5,
                0.98,
                "S",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                0.02,
                "I",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.98,
                0.5,
                "L",
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.02,
                0.5,
                "R",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
        ax.axis("off")

        xy_slice = reorient_slice(image[:, :, xyz[2]], 2)
        if not flat:
            ax = plt.subplot(gs[1, 1])
        else:
            ax = plt.subplot(gs[0, 2])
        im = ax.imshow(xy_slice, cmap=cmap, vmin=vmin, vmax=vmax)
        if overlay is not None:
            xy_overlay = reorient_slice(overlay[:, :, xyz[2]], 2)
            im = ax.imshow(xy_overlay, alpha=overlay_alpha, cmap=overlay_cmap, vmin=vminol, vmax=vmaxol)

        if xyz_lines:
            # add lines
            l = mlines.Line2D(
                [xy_slice.shape[0] - xyz[0], xy_slice.shape[0] - xyz[0]],
                [xyz_pad, xy_slice.shape[0] - xyz_pad],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
            l = mlines.Line2D(
                [xyz_pad, xy_slice.shape[1] - xyz_pad],
                [xy_slice.shape[1] - xyz[1], xy_slice.shape[1] - xyz[1]],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
        if orient_labels:
            ax.text(
                0.5,
                0.98,
                "A",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                0.02,
                "P",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.98,
                0.5,
                "L",
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.02,
                0.5,
                "R",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
        ax.axis("off")

        if not flat:
            # empty corner
            ax = plt.subplot(gs[1, 0])
            if text is not None:
                # add text
                left, width = 0.25, 0.5
                bottom, height = 0.25, 0.5
                right = left + width
                top = bottom + height
                ax.text(
                    0.5 * (left + right) + text_dx,
                    0.5 * (bottom + top) + text_dy,
                    text,
                    horizontalalignment="center",
                    verticalalignment="center",
                    fontsize=textfontsize,
                    color=textfontcolor,
                    transform=ax.transAxes,
                )
            # ax.text(0.5, 0.5)
            ax.imshow(np.zeros(image.shape[:-1]), cmap="Greys_r")
            ax.axis("off")

        if cbar:
            cbar_start = (1 - cbar_length) / 2
            if cbar_vertical:
                cax = fig.add_axes([0.9 + cbar_dx, cbar_start, 0.03, cbar_length])
                cbar_orient = "vertical"
            else:
                cax = fig.add_axes([cbar_start, 0.08 + cbar_dx, cbar_length, 0.03])
                cbar_orient = "horizontal"
            fig.colorbar(im, cax=cax, orientation=cbar_orient)

    ## multi-channel images ##
    elif image.components > 1:
        raise ValueError("Multi-channel images not currently supported!")

    if filename is not None:
        plt.savefig(filename, dpi=dpi, transparent=transparent)
        plt.close(fig)
    else:
        plt.show()

    # turn warnings back to default
    warnings.simplefilter("default")




"""
Test ants_image.py

nptest.assert_allclose
self.assertEqual
self.assertTrue

    def test_initialize_and_get_value(self):
        for img, metric in self.metrics:
            img2 = img.clone() * 1.1
            metric.set_fixed_image(img)
            metric.set_moving_image(img2)
            metric.set_sampling()
            metric.initialize()
            value = metric.get_value()

    def test__call__(self):
        for img, metric in self.metrics:
            img2 = img.clone() * 1.1
            imgmask = img > img.mean()
            imgmask2 = img2 > img2.mean()
            val = metric(img,img2)
            val = metric(img,img2, fixed_mask=imgmask, moving_mask=imgmask2)
            val = metric(img,img2, sampling_percentage=0.8)


"""

import os
import unittest
from common import run_tests

from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants


class TestClass_ANTsImageToImageMetric(unittest.TestCase):
    """
    Test ants.ANTsImage class
    """
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        metric2d = ants.new_ants_metric(precision='float', dimension=2)
        metric3d = ants.new_ants_metric(precision='float', dimension=3)

        self.imgs = [img2d, img3d]
        self.metrics = [metric2d, metric3d]

    def tearDown(self):
        pass

    def test__repr__(self):
        for metric in self.metrics:
            r = metric.__repr__()

    def test_precision(self):
        for metric in self.metrics:
            precison = metric.precision

    def dimension(self):
        for metric in self.metrics:
            dimension = metric.dimension

    def metrictype(self):
        for metric in self.metrics:
            mtype  = metric.metrictype

    def is_vector(self):
        for metric in self.metrics:
            isvec = metric.is_vector

    def pointer(self):
        for metric in self.metrics:
            ptr = metric.pointer

    def test_set_fixed_image(self):
        # def set_fixed_image(self, image):
        for img, metric in zip(self.imgs,self.metrics):
            metric.set_fixed_image(img)

    def test_set_fixed_mask(self):
        # def set_fixed_image(self, image):
        for img, metric in zip(self.imgs,self.metrics):
            mask = img > img.mean()
            metric.set_fixed_mask(img)

    def test_set_moving_image(self):
        # def set_fixed_image(self, image):
        for img, metric in zip(self.imgs,self.metrics):
            metric.set_moving_image(img)

    def test_set_moving_mask(self):
        # def set_fixed_image(self, image):
        for img, metric in zip(self.imgs,self.metrics):
            mask = img > img.mean()
            metric.set_moving_mask(img)

    def test_set_sampling(self):
        for img, metric in zip(self.imgs,self.metrics):

            with self.assertRaises(Exception):
                # set sampling without moving+fixed images
                metric.set_sampling()
            
            img2 = img.clone()
            metric.set_fixed_image(img)
            metric.set_moving_image(img2)
            metric.set_sampling()

    def test_initialize(self):
        for img, metric in zip(self.imgs,self.metrics):

            with self.assertRaises(Exception):
                # initialize without moving+fixed images
                metric.initialize()
            
            img2 = img.clone()
            metric.set_fixed_image(img)
            metric.set_moving_image(img2)
            metric.set_sampling()
            metric.initialize()

    def test_get_value(self):
        for img, metric in zip(self.imgs,self.metrics):

            with self.assertRaises(Exception):
                # initialize without moving+fixed images
                metric.get_value()
            
            img2 = img.clone()
            metric.set_fixed_image(img)
            metric.set_moving_image(img2)
            metric.set_sampling()
            metric.get_value()

    def test__call__(self):
        for img, metric in zip(self.imgs,self.metrics):
            img2 = img.clone() * 1.1
            imgmask = img > img.mean()
            imgmask2 = img2 > img2.mean()
            val = metric(img,img2)
            # val = metric(img,img2, fixed_mask=imgmask, moving_mask=imgmask2)
            val = metric(img,img2, sampling_percentage=0.8)


if __name__ == '__main__':
    run_tests()



"""
Test ants.registration module

nptest.assert_allclose
self.assertEqual
self.assertTrue
"""

import os
import unittest
from common import run_tests
from tempfile import mktemp

import numpy as np
import numpy.testing as nptest
import pandas as pd

import ants


class TestModule_affine_initializer(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        # test ANTsPy/ANTsR example
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r27"))
        txfile = ants.affine_initializer(fi, mi)
        tx = ants.read_transform(txfile)


class TestModule_apply_transforms(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        # test ANTsPy/ANTsR example
        fixed = ants.image_read(ants.get_ants_data("r16"))
        moving = ants.image_read(ants.get_ants_data("r64"))
        fixed = ants.resample_image(fixed, (64, 64), 1, 0)
        moving = ants.resample_image(moving, (128, 128), 1, 0)
        mytx = ants.registration(fixed=fixed, moving=moving, type_of_transform="SyN")
        mywarpedimage = ants.apply_transforms(
            fixed=fixed, moving=moving, transformlist=mytx["fwdtransforms"]
        )
        self.assertEqual(mywarpedimage.pixeltype, moving.pixeltype)
        self.assertTrue(ants.image_physical_space_consistency(fixed, mywarpedimage,
                                                              0.0001, datatype = False))

        # Call with float precision for transforms, but should still return input type
        mywarpedimage2 = ants.apply_transforms(
            fixed=fixed, moving=moving, transformlist=mytx["fwdtransforms"], singleprecision=True
        )
        self.assertEqual(mywarpedimage2.pixeltype, moving.pixeltype)
        self.assertLessEqual(np.sum((mywarpedimage.numpy() - mywarpedimage2.numpy()) ** 2), 0.1)

        # bad interpolator
        with self.assertRaises(Exception):
            mywarpedimage = ants.apply_transforms(
                fixed=fixed,
                moving=moving,
                transformlist=mytx["fwdtransforms"],
                interpolator="unsupported-interp",
            )

        # transform doesnt exist
        with self.assertRaises(Exception):
            mywarpedimage = ants.apply_transforms(
                fixed=fixed,
                moving=moving,
                transformlist=["blah-blah.mat"],
                interpolator="unsupported-interp",
            )


class TestModule_create_jacobian_determinant_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r64"))
        fi = ants.resample_image(fi, (128, 128), 1, 0)
        mi = ants.resample_image(mi, (128, 128), 1, 0)
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform=("SyN"))
        try:
            jac = ants.create_jacobian_determinant_image(
                fi, mytx["fwdtransforms"][0], 1
            )
        except:
            pass


class TestModule_create_warped_grid(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r64"))
        mygr = ants.create_warped_grid(mi)

        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform=("SyN"))
        mywarpedgrid = ants.create_warped_grid(
            mi,
            grid_directions=(False, True),
            transform=mytx["fwdtransforms"],
            fixed_reference_image=fi,
        )


class TestModule_fsl2antstransform(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fslmat = np.zeros((4, 4))
        np.fill_diagonal(fslmat, 1)
        img = ants.image_read(ants.get_ants_data("ch2"))
        tx = ants.fsl2antstransform(fslmat, img, img)


class TestModule_interface(unittest.TestCase):
    def setUp(self):
        self.transform_types = {
            "SyNBold",
            "SyNBoldAff",
            "ElasticSyN",
            "SyN",
            "SyNRA",
            "SyNOnly",
            "SyNAggro",
            "SyNCC",
            "TRSAA",
            "SyNabp",
            "SyNLessAggro",
            "TVMSQ",
            "TVMSQC",
            "Rigid",
            "Similarity",
            "Translation",
            "Affine",
            "AffineFast",
            "BOLDAffine",
            "QuickRigid",
            "DenseRigid",
            "BOLDRigid",
            "antsRegistrationSyNQuick[b,32,26]",
            "antsRegistrationSyNQuick[s]",
            "antsRegistrationSyNRepro[s]",
            "antsRegistrationSyN[s]"
        }

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r64"))
        fi = ants.resample_image(fi, (60, 60), 1, 0)
        mi = ants.resample_image(mi, (60, 60), 1, 0)
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform="SyN")

    def test_affine_interface(self):
        print("Starting affine interface registration test")
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r64"))
        with self.assertRaises(ValueError):
            ants.registration(
                fixed=fi,
                moving=mi,
                type_of_transform="Translation",
                aff_iterations=4,
                aff_shrink_factors=4,
                aff_smoothing_sigmas=(4, 4),
            )

        mytx = ants.registration(
            fixed=fi,
            moving=mi,
            type_of_transform="Affine",
            aff_iterations=(4, 4),
            aff_shrink_factors=(4, 4),
            aff_smoothing_sigmas=(4, 4),
        )
        mytx = ants.registration(
            fixed=fi,
            moving=mi,
            type_of_transform="Translation",
            aff_iterations=4,
            aff_shrink_factors=4,
            aff_smoothing_sigmas=4,
        )

    def test_registration_types(self):
        print("Starting long registration interface test")
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r64"))
        fi = ants.resample_image(fi, (60, 60), 1, 0)
        mi = ants.resample_image(mi, (60, 60), 1, 0)

        for ttype in self.transform_types:
            print(ttype)
            mytx = ants.registration(fixed=fi, moving=mi, type_of_transform=ttype)

            # with mask
            fimask = fi > fi.mean()
            mytx = ants.registration(
                fixed=fi, moving=mi, mask=fimask, type_of_transform=ttype
            )
        print("Finished long registration interface test")

    def test_reg_precision_option(self):
        # Check that registration works with float and double precision
        fi = ants.image_read(ants.get_ants_data("r16"))
        mi = ants.image_read(ants.get_ants_data("r64"))
        fi = ants.resample_image(fi, (60, 60), 1, 0)
        mi = ants.resample_image(mi, (60, 60), 1, 0)
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform="SyN") # should be float precision
        info = ants.image_header_info(mytx["fwdtransforms"][0])
        self.assertEqual(info['pixeltype'], 'float')
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform="SyN", singleprecision=False)
        info = ants.image_header_info(mytx["fwdtransforms"][0])
        self.assertEqual(info['pixeltype'], 'double')


class TestModule_metrics(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data("r16")).clone("float")
        mi = ants.image_read(ants.get_ants_data("r64")).clone("float")
        mival = ants.image_mutual_information(fi, mi)  # -0.1796141


class TestModule_reflect_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        axis = 2
        asym = ants.reflect_image(fi, axis, "Affine")["warpedmovout"]
        asym = asym - fi


class TestModule_reorient_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_reorient_image(self):
        mni = ants.image_read(ants.get_data('mni'))
        mni2 = mni.reorient_image2()

    def test_get_center_of_mass(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        com = ants.get_center_of_mass(fi)

        self.assertEqual(len(com), fi.dimension)

        fi = ants.image_read(ants.get_ants_data("r64"))
        com = ants.get_center_of_mass(fi)
        self.assertEqual(len(com), fi.dimension)

        fi = fi.clone("unsigned int")
        com = ants.get_center_of_mass(fi)
        self.assertEqual(len(com), fi.dimension)

        # 3d
        img = ants.image_read(ants.get_ants_data("mni"))
        com = ants.get_center_of_mass(img)
        self.assertEqual(len(com), img.dimension)


class TestModule_resample_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_resample_image_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        finn = ants.resample_image(fi, (50, 60), True, 0)
        filin = ants.resample_image(fi, (1.5, 1.5), False, 1)

    def test_resample_channels(self):
        img = ants.image_read( ants.get_ants_data("r16"))
        img = ants.merge_channels([img, img])
        outimg = ants.resample_image(img, (128,128), True)
        self.assertEqual(outimg.shape, (128, 128))
        self.assertEqual(outimg.components, 2)

    def test_resample_image_to_target_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        fi2mm = ants.resample_image(fi, (2, 2), use_voxels=0, interp_type=1)
        resampled = ants.resample_image_to_target(fi2mm, fi, verbose=True)
        self.assertTrue(ants.image_physical_space_consistency(fi, resampled, 0.0001, datatype=True))


class TestModule_symmetrize_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        simage = ants.symmetrize_image(image)


class TestModule_build_template(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r27"))
        timage = ants.build_template(image_list=(image, image2))

    def test_type_of_transform(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r27"))
        timage = ants.build_template(image_list=(image, image2))
        timage = ants.build_template(
            image_list=(image, image2), type_of_transform="SyNCC"
        )


class TestModule_multivar(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r27"))
        demonsMetric = ["demons", image, image2, 1, 1]
        ccMetric = ["CC", image, image2, 2, 1]
        metrics = list()
        metrics.append(demonsMetric)
        reg3 = ants.registration(image, image2, "SyNOnly", multivariate_extras=metrics)
        metrics.append(ccMetric)
        reg2 = ants.registration(
            image, image2, "SyNOnly", multivariate_extras=metrics, verbose=True
        )

class TestModule_random(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_landmark_transforms(self):
        fixed = np.array([[50.0,50.0],[200.0,50.0],[200.0,200.0]])
        moving = np.array([[50.0,50.0],[50.0,200.0],[200.0,200.0]])
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="syn",
                                            domain_image=ants.image_read(ants.get_data('r16')),
                                            verbose=True)
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="tv",
                                            domain_image=ants.image_read(ants.get_data('r16')))
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="affine")
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="rigid")
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="similarity")
        domain_image = ants.image_read(ants.get_ants_data("r16"))
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="bspline", domain_image=domain_image, number_of_fitting_levels=5)
        xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="diffeo", domain_image=domain_image, number_of_fitting_levels=6)

        res = ants.fit_time_varying_transform_to_point_sets([fixed, moving, moving],
                                                            domain_image=ants.image_read(ants.get_data('r16')),
                                                            verbose=True)

    def test_deformation_gradient(self):
        fi = ants.image_read( ants.get_ants_data('r16'))
        mi = ants.image_read( ants.get_ants_data('r64'))
        fi = ants.resample_image(fi,(128,128),1,0)
        mi = ants.resample_image(mi,(128,128),1,0)
        mytx = ants.registration(fixed=fi , moving=mi, type_of_transform = ('SyN') )

        dg = ants.deformation_gradient( ants.image_read( mytx['fwdtransforms'][0] ) )
        # Expect some differences between these two methods
        dg_py = ants.deformation_gradient( ants.image_read( mytx['fwdtransforms'][0] ),
                                       py_based=True)

        rot = ants.deformation_gradient( ants.image_read( mytx['fwdtransforms'][0] ),
                                       to_rotation=True)
        rot_py = ants.deformation_gradient( ants.image_read( mytx['fwdtransforms'][0] ),
                                       to_rotation=True, py_based=True)

        def rotation_angle_diff_field(rot_1, rot_2, degrees=False):
            """
            Compute angular difference between two rotation matrix images.
            Inputs:
                rot_1: antsImage with dim*dim components representing the matrix
                rot_2: antsImage with same shape and dimension as rot_1
            Returns:
                angle_diff: numpy array with angle in radians (or degrees if degrees=True)
            """
            # cast and reshape
            dim = rot_1.dimension
            if rot_2.dimension != dim:
                raise ValueError("Rotation images must have the same dimension.")
            if rot_1.shape != rot_2.shape:
                raise ValueError("Rotation images must have the same shape.")
            rot_py1 = rot_1.numpy().reshape(rot_1.shape + (dim, dim))
            rot_py2 = rot_2.numpy().reshape(rot_2.shape + (dim, dim))
            # Compute relative rotation: R_rel = R_py @ R.T
            R_rel = rot_py1 @ np.swapaxes(rot_py2, -2, -1)
            trace = np.trace(R_rel, axis1=-2, axis2=-1)
            if dim == 2:
                angle = np.arccos(np.clip((trace) / 2, -1.0, 1.0))
            elif dim == 3:
                angle = np.arccos(np.clip((trace - 1) / 2, -1.0, 1.0))
            else:
                raise ValueError("Only 2D or 3D rotation matrices are supported.")

            return np.degrees(angle) if degrees else angle

        rot_angle_diff = rotation_angle_diff_field(rot_py, rot, degrees=True)
        self.assertTrue(np.all(rot_angle_diff < 1))

        rot_inv = ants.deformation_gradient( ants.image_read( mytx['fwdtransforms'][0] ),
                                       to_inverse_rotation=True)
        rot_py_inv = ants.deformation_gradient( ants.image_read( mytx['fwdtransforms'][0] ),
                                       to_inverse_rotation=True, py_based=True)

        rot_angle_diff = rotation_angle_diff_field(rot_inv, rot_py_inv, degrees=True)
        self.assertTrue(np.all(rot_angle_diff < 1))

        # Check it's actually the inverse
        self.assertTrue(np.allclose (rot_py.numpy(), rot_py_inv.numpy()[..., [0, 2, 1, 3]]))

    def test_jacobian(self):
        fi = ants.image_read( ants.get_ants_data('r16'))
        mi = ants.image_read( ants.get_ants_data('r64'))
        fi = ants.resample_image(fi,(128,128),1,0)
        mi = ants.resample_image(mi,(128,128),1,0)
        mytx = ants.registration(fixed=fi , moving=mi, type_of_transform = ('SyN') )
        jac = ants.create_jacobian_determinant_image(fi,mytx['fwdtransforms'][0],1)

    def test_apply_transforms(self):
        fixed = ants.image_read( ants.get_ants_data('r16') )
        moving = ants.image_read( ants.get_ants_data('r64') )
        fixed = ants.resample_image(fixed, (64,64), 1, 0)
        moving = ants.resample_image(moving, (64,64), 1, 0)
        mytx = ants.registration(fixed=fixed , moving=moving ,
                                type_of_transform = 'SyN' )
        mywarpedimage = ants.apply_transforms( fixed=fixed, moving=moving,
                                            transformlist=mytx['fwdtransforms'] )

    def test_apply_transforms_to_points(self):
        fixed = ants.image_read( ants.get_ants_data('r16') )
        moving = ants.image_read( ants.get_ants_data('r27') )
        reg = ants.registration( fixed, moving, 'Affine' )
        d = {'x': [128, 127], 'y': [101, 111]}
        pts = pd.DataFrame(data=d)
        ptsw = ants.apply_transforms_to_points( 2, pts, reg['fwdtransforms'])

    def test_warped_grid(self):
        fi = ants.image_read( ants.get_ants_data( 'r16' ) )
        mi = ants.image_read( ants.get_ants_data( 'r64' ) )
        mygr = ants.create_warped_grid( mi )
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = ('SyN') )
        mywarpedgrid = ants.create_warped_grid( mi, grid_directions=(False,True),
                            transform=mytx['fwdtransforms'], fixed_reference_image=fi )

    def test_more_registration(self):
        fi = ants.image_read(ants.get_ants_data('r16'))
        mi = ants.image_read(ants.get_ants_data('r64'))
        fi = ants.resample_image(fi, (60,60), 1, 0)
        mi = ants.resample_image(mi, (60,60), 1, 0)
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'SyN' )
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'antsRegistrationSyN[t]' )
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'antsRegistrationSyN[b]' )
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'antsRegistrationSyN[s]' )

    def test_motion_correction(self):
        fi = ants.image_read(ants.get_ants_data('ch2'))
        mytx = ants.motion_correction( fi )

    def test_label_image_registration(self):
        fi = ants.image_read(ants.get_ants_data('r16'))
        mi = ants.image_read(ants.get_ants_data('r64'))
        fi = ants.resample_image(fi, (60,60), 1, 0)
        mi = ants.resample_image(mi, (60,60), 1, 0)
        fi_seg = ants.threshold_image(fi, "Kmeans", 3)-1
        mi_seg = ants.threshold_image(mi, "Kmeans", 3)-1
        mytx = ants.label_image_registration([fi_seg],
                                             [mi_seg],
                                             fixed_intensity_images=fi,
                                             moving_intensity_images=mi)


if __name__ == "__main__":
    run_tests()



"""
Test ants_image.py

nptest.assert_allclose
self.assertEqual
self.assertTrue

"""

import os
import unittest
from common import run_tests

from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants


class TestModule_ants_transform_io(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        tx2d = ants.create_ants_transform(precision='float', dimension=2)
        tx3d = ants.create_ants_transform(precision='float', dimension=3)

        self.imgs = [img2d, img3d]
        self.txs = [tx2d, tx3d]
        self.pixeltypes = ['unsigned char', 'unsigned int', 'float']

        self.matrix_offset_types = ['AffineTransform',
                                    'CenteredAffineTransform',
                                    'Euler2DTransform',
                                    'Euler3DTransform',
                                    'Rigid2DTransform',
                                    'QuaternionRigidTransform',
                                    'Similarity2DTransform',
                                    'CenteredSimilarity2DTransform',
                                    'Similarity3DTransform',
                                    'CenteredRigid2DTransform',
                                    'CenteredEuler3DTransform',
                                    'Rigid3DTransform']

    def tearDown(self):
        pass

    def test_new_ants_transform(self):
        tx = ants.new_ants_transform()
        params = tx.parameters*2

        # initialize w/ params
        tx = ants.new_ants_transform(parameters=params)

    def test_create_ants_transform(self):
        for mtype in self.matrix_offset_types:
            for ptype in {'float', 'double'}:
                for ndim in {2, 3}:
                    ants.create_ants_transform(transform_type=mtype, precision=ptype, dimension=ndim)

        # supported types
        tx = ants.create_ants_transform(supported_types=True)

        # input params
        translation = (3,4,5)
        tx = ants.create_ants_transform( transform_type='Euler3DTransform', translation=translation )

        #translation = np.array([3,4,5])
        tx = ants.create_ants_transform( transform_type='Euler3DTransform', translation=translation )

        # invalid dimension
        with self.assertRaises(Exception):
            ants.create_ants_transform(dimension=5)

        # unsupported type
        with self.assertRaises(Exception):
            ants.create_ants_transform(transform_type='unsupported-type')

        # bad param arg
        with self.assertRaises(Exception):
            ants.create_ants_transform( transform_type='Euler3DTransform', translation=ants.image_read(ants.get_ants_data('r16')) )

        # bad precision
        with self.assertRaises(Exception):
            ants.create_ants_transform(precision='unsigned int')

    def test_read_write_transform(self):
        for tx in self.txs:
            filename = mktemp(suffix='.mat')
            ants.write_transform(tx, filename)
            tx2 = ants.read_transform(filename, precision='float')

        # file doesnt exist
        with self.assertRaises(Exception):
            ants.read_transform('blah-blah.mat')

    def test_from_displacement_components(self):
        vec_np = np.ndarray((2,2,3), dtype=np.float32)
        vec = ants.from_numpy(vec_np, origin=(0,0), spacing=(1,1), has_components=True)
        # should get ValueError here because the 2D vector field has 3 components
        with self.assertRaises(ValueError):
            ants.transform_from_displacement_field(vec)
        vec_np = np.ndarray((2,2,2,3), dtype=np.float32)
        vec = ants.from_numpy(vec_np, origin=(0,0,0), spacing=(1,1,1), has_components=True)
        # should work here because the 3D vector field has 3 components
        tx = ants.transform_from_displacement_field(vec)

    def test_from_displacement(self):
        fi = ants.image_read(ants.get_ants_data('r16') )
        mi = ants.image_read(ants.get_ants_data('r64') )
        fi = ants.resample_image(fi,(60,60),1,0)
        mi = ants.resample_image(mi,(60,60),1,0) # speed up
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = ('SyN') )
        # read transform, which calls transform_from_displacement_field
        atx = ants.read_transform( mytx['fwdtransforms'][0] )

    def test_to_displacement(self):
        fi = ants.image_read(ants.get_ants_data('r16') )
        mi = ants.image_read(ants.get_ants_data('r64') )
        fi = ants.resample_image(fi,(60,60),1,0)
        mi = ants.resample_image(mi,(60,60),1,0) # speed up
        mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = ('SyN') )
        vec = ants.image_read( mytx['fwdtransforms'][0] )
        atx = ants.transform_from_displacement_field( vec )
        field = ants.transform_to_displacement_field( atx, fi )

    def test_catch_error(self):
        with self.assertRaises(Exception):
            ants.write_transform(123, 'test.mat')


if __name__ == '__main__':
    run_tests()



__all__ = ['histogram_equalize_image']

import numpy as np

import ants
from ants.decorators import image_method

@image_method
def histogram_equalize_image(image, number_of_histogram_bins=256):
    """
    Histogram equalize image

    # from http://www.janeriksolem.net/histogram-equalization-with-python-and.html      
    
    Arguments
    ---------
    image : ANTsImage
        source image

    number_of_histogram_bins : integer
        number of bins for cumulative histogram

    Example
    -------
    >>> import ants
    >>> src_img = ants.image_read(ants.get_data('r16'))
    >>> src_img_eq = ants.histogram_equalize_image(src_img)
    """

    image_array = image.numpy()
    image_histogram, bins = np.histogram(image_array.flatten(), number_of_histogram_bins, density=True)
    cdf = image_histogram.cumsum()
    cdf = image_array.max() * cdf / cdf[-1]
    image_array_equalized_flat = np.interp(image_array.flatten(), bins[:-1], cdf)
    image_array_equalized = image_array_equalized_flat.reshape(image_array.shape)    
    image_array_equalized = ( ( image_array_equalized - image_array_equalized.min() ) /
                              ( image_array_equalized.max() - image_array_equalized.min() ) )
    image_array_equalized = image_array_equalized * ( image_array.max() - image_array.min() ) + image_array.min()

    image_equalized = ants.from_numpy(image_array_equalized)

    return ants.copy_image_info(image, image_equalized)





import numpy as np
import ants
from ants.decorators import image_method

__all__ = ['image_physical_space_consistency',
           'allclose']

@image_method
def image_physical_space_consistency(image1, image2, tolerance=1e-2, datatype=False):
    """
    Check if two or more ANTsImage objects occupy the same physical space

    ANTsR function: `antsImagePhysicalSpaceConsistency`

    Arguments
    ---------
    *images : ANTsImages
        images to compare

    tolerance : float
        tolerance when checking origin and spacing

    data_type : boolean
        If true, also check that the image data types are the same

    Returns
    -------
    boolean
        true if images share same physical space, false otherwise
    """
    images = [image1, image2]

    img1 = images[0]
    for img2 in images[1:]:
        if (not ants.is_image(img1)) or (not ants.is_image(img2)):
            raise ValueError('Both images must be of class `AntsImage`')

        # image dimension check
        if img1.dimension != img2.dimension:
            return False

        # image spacing check
        space_diffs = sum([abs(s1-s2)>tolerance for s1, s2 in zip(img1.spacing, img2.spacing)])
        if space_diffs > 0:
            return False

        # image origin check
        origin_diffs = sum([abs(s1-s2)>tolerance for s1, s2 in zip(img1.origin, img2.origin)])
        if origin_diffs > 0:
            return False

        # image direction check
        origin_diff = np.allclose(img1.direction, img2.direction, atol=tolerance)
        if not origin_diff:
            return False

        # data type
        if datatype == True:
            if img1.pixeltype != img2.pixeltype:
                return False

            if img1.components != img2.components:
                return False

    return True


@image_method
def allclose(image1, image2):
    """
    Check if two images have the same array values
    """
    return np.allclose(image1.numpy(), image2.numpy())



__all__ = ['histogram_match_image',
           'histogram_match_image2']

import numpy as np

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn


@image_method
def histogram_match_image(source_image, reference_image, number_of_histogram_bins=255, number_of_match_points=64, use_threshold_at_mean_intensity=False):
    """
    Histogram match source image to reference image.

    Arguments
    ---------
    source_image : ANTsImage
        source image

    reference_image : ANTsImage
        reference image

    number_of_histogram_bins : integer
        number of bins for source and reference histograms

    number_of_match_points : integer
        number of points for histogram matching

    use_threshold_at_mean_intensity : boolean
        see ITK description.

    Example
    -------
    >>> import ants
    >>> src_img = ants.image_read(ants.get_data('r16'))
    >>> ref_img = ants.image_read(ants.get_data('r64'))
    >>> src_ref = ants.histogram_match_image(src_img, ref_img)
    """

    inpixeltype = source_image.pixeltype
    ndim = source_image.dimension
    if source_image.pixeltype != 'float':
        source_image = source_image.clone('float')
    if reference_image.pixeltype != 'float':
        reference_image = reference_image.clone('float')

    libfn = get_lib_fn('histogramMatchImageF%i' % ndim)
    itkimage = libfn(source_image.pointer, reference_image.pointer, number_of_histogram_bins, number_of_match_points, use_threshold_at_mean_intensity)

    new_image = ants.from_pointer(itkimage).clone(inpixeltype)
    return new_image

@image_method
def histogram_match_image2(source_image, reference_image, 
                           source_mask=None, reference_mask=None,
                           match_points=64,
                           transform_domain_size=255):
    """
    Transform image intensities based on histogram mapping.

    Apply B-spline 1-D maps to an input image for intensity warping.

    Arguments
    ---------
    source_image : ANTsImage
        source image

    reference_image : ANTsImage
        reference image

    source_mask : ANTsImage
        source mask

    reference_mask : ANTsImage
        reference mask
                
    match_points : integer or tuple
        Parametric points at which the intensity transform displacements are 
        specified between [0, 1], i.e. quantiles.  Alternatively, a single number 
        can be given and the sequence is linearly spaced in [0, 1]. 

    transform_domain_size : integer
        Defines the sampling resolution of the B-spline warping.

    Returns
    -------
    ANTs image

    Example
    -------
    >>> import ants
    >>> src_img = ants.image_read(ants.get_data('r16'))
    >>> ref_img = ants.image_read(ants.get_data('r64'))
    >>> src_ref = ants.histogram_match_image2(src_img, ref_img)
    """

    if not isinstance(match_points, int):
        if any(b < 0 for b in match_points) and any(b > 1 for b in match_points):
            raise ValueError("If specifying match_points as a vector, values must be in the range [0, 1]")

    # Use entire image if mask isn't specified
    if source_mask is None:
        source_mask = source_image * 0 + 1
    if reference_mask is None:
        reference_mask = reference_image * 0 + 1

    source_array = source_image.numpy()
    source_mask_array = source_mask.numpy()
    source_masked_min = source_image[source_mask != 0].min()
    source_masked_max = source_image[source_mask != 0].max()

    reference_array = reference_image.numpy()
    reference_mask_array = reference_mask.numpy()

    parametric_points = None
    if not isinstance(match_points, int):
        parametric_points = match_points
    else:
        parametric_points = np.linspace(0, 1, match_points)

    source_intensity_quantiles = np.quantile(source_array[source_mask_array != 0], parametric_points)
    reference_intensity_quantiles = np.quantile(reference_array[reference_mask_array != 0], parametric_points)
    displacements = reference_intensity_quantiles - source_intensity_quantiles

    scattered_data = np.reshape(displacements, (len(displacements), 1))
    parametric_data = np.reshape(parametric_points * (source_masked_max - source_masked_min) + source_masked_min, (len(parametric_points), 1))

    transform_domain_origin = source_masked_min
    transform_domain_spacing = (source_masked_max - transform_domain_origin) / (transform_domain_size - 1)

    bspline_histogram_transform = ants.fit_bspline_object_to_scattered_data(scattered_data,
        parametric_data, [transform_domain_origin], [transform_domain_spacing], [transform_domain_size],
        data_weights=None, is_parametric_dimension_closed=None, number_of_fitting_levels=8, 
        mesh_size=1, spline_order=3)

    transform_domain = np.linspace(source_masked_min, source_masked_max, transform_domain_size)

    transformed_source_array = source_image.numpy()
    for i in range(len(transform_domain) - 1):
        indices = np.where((source_array >= transform_domain[i]) & (source_array < transform_domain[i+1]))
        intensities = source_array[indices]

        alpha = (intensities - transform_domain[i])/(transform_domain[i+1] - transform_domain[i])
        xfrm = alpha * (bspline_histogram_transform[i+1] - bspline_histogram_transform[i]) + bspline_histogram_transform[i]
        transformed_source_array[indices] = intensities + xfrm

    transformed_source_image = ants.from_numpy(transformed_source_array, origin=source_image.origin,
        spacing=source_image.spacing, direction=source_image.direction)
    transformed_source_image[source_mask == 0] = source_image[source_mask == 0]

    return(transformed_source_image)




"""
Test ants_image.py

nptest.assert_allclose
self.assertEqual
self.assertTrue

"""

import os
import unittest
from common import run_tests

from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants


class TestClass_ANTsImage(unittest.TestCase):
    """
    Test ants.ANTsImage class
    """
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        img3d = ants.image_read(ants.get_ants_data('mni'))
        self.imgs = [img2d, img3d]
        self.pixeltypes = ['unsigned char', 'unsigned int', 'float', 'double']
        self.numpy_pixeltypes = ['uint8', 'uint32', 'float32', 'float64']

    def tearDown(self):
        pass

    def test_get_spacing(self):
        #self.setUp()
        for img in self.imgs:
            spacing = img.spacing
            self.assertTrue(isinstance(spacing, tuple))
            self.assertEqual(len(img.spacing), img.dimension)

    def test_set_spacing(self):
        #self.setUp()
        for img in self.imgs:
            # set spacing from list
            new_spacing_list = [6.9]*img.dimension
            img.set_spacing(new_spacing_list)
            self.assertEqual(img.spacing, tuple(new_spacing_list))

            # set spacing from tuple
            new_spacing_tuple = tuple(new_spacing_list)
            img.set_spacing(new_spacing_tuple)
            self.assertEqual(img.spacing, new_spacing_tuple)

            # test exceptions
            with self.assertRaises(Exception):
                new_spacing = img.clone()
                img.set_spacing(new_spacing)
            with self.assertRaises(Exception):
                new_spacing = [6.9]*(img.dimension+1)
                img.set_spacing(new_spacing)

    def test_get_origin(self):
        #self.setUp()
        for img in self.imgs:
            origin = img.origin
            self.assertTrue(isinstance(origin, tuple))
            self.assertEqual(len(img.origin), img.dimension)

    def test_set_origin(self):
        for img in self.imgs:
            # set spacing from list
            new_origin_list = [6.9]*img.dimension
            img.set_origin(new_origin_list)
            self.assertEqual(img.origin, tuple(new_origin_list))

            # set spacing from tuple
            new_origin_tuple = tuple(new_origin_list)
            img.set_origin(new_origin_tuple)
            self.assertEqual(img.origin, new_origin_tuple)

            # test exceptions
            with self.assertRaises(Exception):
                new_origin = img.clone()
                img.set_origin(new_origin)
            with self.assertRaises(Exception):
                new_origin = [6.9]*(img.dimension+1)
                img.set_origin(new_origin)

    def test_get_direction(self):
        #self.setUp()
        for img in self.imgs:
            direction = img.direction
            self.assertTrue(isinstance(direction,np.ndarray))
            self.assertTrue(img.direction.shape, (img.dimension,img.dimension))

    def test_set_direction(self):
        #self.setUp()
        for img in self.imgs:
            new_direction = np.eye(img.dimension)*3
            img.set_direction(new_direction)
            nptest.assert_allclose(img.direction, new_direction)

            # set from list
            img.set_direction(new_direction.tolist())

            # test exceptions
            with self.assertRaises(Exception):
                new_direction = img.clone()
                img.set_direction(new_direction)
            with self.assertRaises(Exception):
                new_direction = np.eye(img.dimension+2)*3
                img.set_direction(new_direction)

    def test_view(self):
        #self.setUp()
        for img in self.imgs:
            myview = img.view()
            self.assertTrue(isinstance(myview, np.ndarray))
            self.assertEqual(myview.shape, img.shape)

            # test that changes to view are direct changes to the image
            myview *= 3.
            nptest.assert_allclose(myview, img.numpy())

    def test_numpy(self):
        #self.setUp()
        for img in self.imgs:
            mynumpy = img.numpy()
            self.assertTrue(isinstance(mynumpy, np.ndarray))
            self.assertEqual(mynumpy.shape, img.shape)

            # test that changes to view are NOT direct changes to the image
            mynumpy *= 3.
            nptest.assert_allclose(mynumpy, img.numpy()*3.)

    def test_clone(self):
        #self.setUp()
        for img in self.imgs:
            orig_ptype = img.pixeltype
            for ptype in [*self.pixeltypes, *self.numpy_pixeltypes]:
                imgclone = img.clone(ptype)

                self.assertIn(ptype, [imgclone.dtype, imgclone.pixeltype])
                self.assertEqual(img.pixeltype, orig_ptype)
                # test physical space consistency
                self.assertTrue(ants.image_physical_space_consistency(img, imgclone))
                if ptype == orig_ptype:
                    # test that they dont share image pointer
                    view1 = img.view()
                    view1 *= 6.9
                    nptest.assert_allclose(view1, imgclone.numpy()*6.9)

    def test_new_image_like(self):
        #self.setUp()
        for img in self.imgs:
            myarray = img.numpy()
            myarray *= 6.9
            imgnew = img.new_image_like(myarray)
            # test physical space consistency
            self.assertTrue(ants.image_physical_space_consistency(img, imgnew))
            # test data
            nptest.assert_allclose(myarray, imgnew.numpy())
            nptest.assert_allclose(myarray, img.numpy()*6.9)

            # test exceptions
            with self.assertRaises(Exception):
                # not ndarray
                new_data = img.clone()
                img.new_image_like(new_data)
            with self.assertRaises(Exception):
                # wrong shape
                new_data = np.random.randn(69,12,21).astype('float32')
                img.new_image_like(new_data)

        with self.assertRaises(Exception):
            # wrong shape with components
            vecimg = ants.from_numpy(np.random.randn(69,12,3).astype('float32'), has_components=True)
            new_data = np.random.randn(69,12,4).astype('float32')
            vecimg.new_image_like(new_data)

    def test_from_numpy_like(self):
        img = ants.image_read(ants.get_data('mni'))

        arr = img.numpy()
        arr *= 2
        img2 = ants.from_numpy_like(arr, img)
        self.assertTrue(ants.image_physical_space_consistency(img, img2))
        self.assertEqual(img2.mean() / img.mean(), 2)

    def test_to_file(self):
        #self.setUp()
        for img in self.imgs:
            filename = mktemp(suffix='.nii.gz')
            img.to_file(filename)
            # test that file now exists
            self.assertTrue(os.path.exists(filename))
            img2 = ants.image_read(filename)
            # test physical space and data
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img.numpy(), img2.numpy())

            try:
                os.remove(filename)
            except:
                pass

    def test_apply(self):
        #self.setUp()
        for img in self.imgs:
            img2 = img.apply(lambda x: x*6.9)
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()*6.9)

    def test_numpy_ops(self):
        for img in self.imgs:
            median = img.median()
            std = img.std()
            argmin = img.argmin()
            argmax = img.argmax()
            flat = img.flatten()
            nonzero = img.nonzero()
            unique = img.unique()

    def test__add__(self):
        #self.setUp()
        for img in self.imgs:
            # op on constant
            img2 = img + 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()+6.9)

            # op on another image
            img2 = img + img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()+img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img + img2

    def test__radd__(self):
        #self.setUp()
        for img in self.imgs:
            # op on constant
            img2 = img.__radd__(6.9)
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy() + 6.9)

            # op on another image
            img2 = img + img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()+img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img + img2

    def test__sub__(self):
        #self.setUp()
        for img in self.imgs:
            # op on constant
            img2 = img - 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()-6.9)

            # op on another image
            img2 = img - img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()-img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img - img2

    def test__rsub__(self):
        #self.setUp()
        for img in self.imgs:
            # op on constant
            img2 = img.__rsub__(6.9)
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), 6.9 - img.numpy())

            # op on another image
            img2 = img - img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()-img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img - img2

    def test__mul__(self):
        #self.setUp()
        for img in self.imgs:
            # op on constant
            img2 = img * 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()*6.9)

            # op on another image
            img2 = img * img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()*img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img * img2

    def test__rmul__(self):
        #self.setUp()
        for img in self.imgs:
            # op on constant
            img2 = img.__rmul__(6.9)
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), 6.9*img.numpy())

            # op on another image
            img2 = img * img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()*img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img * img2

    def test__div__(self):
        #self.setUp()
        for img in self.imgs:
            img = img + 10
            # op on constant
            img2 = img / 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()/6.9)

            # op on another image
            img2 = img / img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()/img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img / img2

    def test__pow__(self):
        #self.setUp()
        for img in self.imgs:
            img = img + 10
            # op on constant
            img2 = img ** 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()**6.9)

            # op on another image
            img2 = img ** img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()**img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img ** img2

    def test__gt__(self):
        #self.setUp()
        for img in self.imgs:
            img2 = img > 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), (img.numpy()>6.9).astype('int'))

            # op on another image
            img2 = img > img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()>img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img > img2

    def test__ge__(self):
        #self.setUp()
        for img in self.imgs:
            img2 = img >= 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), (img.numpy()>=6.9).astype('int'))

            # op on another image
            img2 = img >= img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()>=img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img >= img2

    def test__lt__(self):
        #self.setUp()
        for img in self.imgs:
            img2 = img < 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), (img.numpy()<6.9).astype('int'))

            # op on another image
            img2 = img < img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()<img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img < img2

    def test__le__(self):
        #self.setUp()
        for img in self.imgs:
            img2 = img <= 6.9
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), (img.numpy()<=6.9).astype('int'))

            # op on another image
            img2 = img <= img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()<=img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img <= img2

    def test__eq__(self):
        #self.setUp()
        for img in self.imgs:
            img2 = (img == 6.9)
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), (img.numpy()==6.9).astype('int'))

            # op on another image
            img2 = img == img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()==img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img == img2

    def test__ne__(self):
        #self.setUp()
        for img in self.imgs:
            img2 = (img != 6.9)
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), (img.numpy()!=6.9).astype('int'))

            # op on another image
            img2 = img != img.clone()
            self.assertTrue(ants.image_physical_space_consistency(img, img2))
            nptest.assert_allclose(img2.numpy(), img.numpy()!=img.numpy())

            with self.assertRaises(Exception):
                # different physical space
                img2 = img.clone()
                img2.set_spacing([2.31]*img.dimension)
                img3 = img != img2

    def test__getitem__(self):
        for img in self.imgs:
            if img.dimension == 2:
                img2 = img[6:9,6:9]
                nptest.assert_allclose(img2.numpy(), img.numpy()[6:9,6:9])
            elif img.dimension == 3:
                img2 = img[6:9,6:9,6:9]
                nptest.assert_allclose(img2.numpy(), img.numpy()[6:9,6:9,6:9])

            # get from another image
            img2 = img.clone()
            xx = img[img2]
            with self.assertRaises(Exception):
                # different physical space
                img2.set_direction(img.direction*2)
                xx = img[img2]

    def test__setitem__(self):
        for img in self.imgs:
            if img.dimension == 2:
                img[6:9,6:9] = 6.9
                self.assertTrue(img.numpy()[6:9,6:9].mean(), 6.9)
            elif img.dimension == 3:
                img[6:9,6:9,6:9] = 6.9
                self.assertTrue(img.numpy()[6:9,6:9,6:9].mean(), 6.9)

            # set from another image
            img2 = img.clone()
            img3 = img.clone()
            img[img2] = 0
            with self.assertRaises(Exception):
                # different physical space
                img2.set_direction(img3.direction*2)
                img3[img2] = 0

    def test__repr__(self):
        for img in self.imgs:
            s = img.__repr__()


    def test__iter__(self):
        # Attempts to use __iter__ on an ANTsImage should raise a TypeError
        img = self.imgs[0]
        with self.assertRaises(TypeError):
            for _ in img:
                pass

class TestModule_ants_image(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16')).clone('float')
        img3d = ants.image_read(ants.get_ants_data('mni')).clone('float')
        self.imgs = [img2d, img3d]
        self.pixeltypes = ['unsigned char', 'unsigned int', 'float', 'double']

    def tearDown(self):
        pass

    def test_copy_image_info(self):
        for img in self.imgs:
            img2 = img.clone()
            img2.set_spacing([6.9]*img.dimension)
            img2.set_origin([6.9]*img.dimension)
            self.assertTrue(not ants.image_physical_space_consistency(img,img2))

            img3 = ants.copy_image_info(reference=img, target=img2)
            self.assertTrue(ants.image_physical_space_consistency(img,img3))

    def test_get_spacing(self):
        for img in self.imgs:
            spacing = ants.get_spacing(img)
            self.assertTrue(isinstance(spacing, tuple))
            self.assertEqual(len(ants.get_spacing(img)), img.dimension)

    def test_set_spacing(self):
        for img in self.imgs:
            # set spacing from list
            new_spacing_list = [6.9]*img.dimension
            ants.set_spacing(img, new_spacing_list)
            self.assertEqual(img.spacing, tuple(new_spacing_list))

            # set spacing from tuple
            new_spacing_tuple = tuple(new_spacing_list)
            ants.set_spacing(img, new_spacing_tuple)
            self.assertEqual(ants.get_spacing(img), new_spacing_tuple)

    def test_get_origin(self):
        for img in self.imgs:
            origin = ants.get_origin(img)
            self.assertTrue(isinstance(origin, tuple))
            self.assertEqual(len(ants.get_origin(img)), img.dimension)

    def test_set_origin(self):
        for img in self.imgs:
            # set spacing from list
            new_origin_list = [6.9]*img.dimension
            ants.set_origin(img, new_origin_list)
            self.assertEqual(img.origin, tuple(new_origin_list))

            # set spacing from tuple
            new_origin_tuple = tuple(new_origin_list)
            ants.set_origin(img, new_origin_tuple)
            self.assertEqual(ants.get_origin(img), new_origin_tuple)

    def test_get_direction(self):
        for img in self.imgs:
            direction = ants.get_direction(img)
            self.assertTrue(isinstance(direction,np.ndarray))
            self.assertTrue(ants.get_direction(img).shape, (img.dimension,img.dimension))

    def test_set_direction(self):
        for img in self.imgs:
            new_direction = np.eye(img.dimension)*3
            ants.set_direction(img, new_direction)
            nptest.assert_allclose(ants.get_direction(img), new_direction)

    def test_image_physical_spacing_consistency(self):
        for img in self.imgs:
            self.assertTrue(ants.image_physical_space_consistency(img,img))
            self.assertTrue(ants.image_physical_space_consistency(img,img,datatype=True))
            clonetype = 'float' if img.pixeltype != 'float' else 'unsigned int'
            img2 = img.clone(clonetype)
            self.assertTrue(ants.image_physical_space_consistency(img,img2))
            self.assertTrue(not ants.image_physical_space_consistency(img,img2,datatype=True))

            # test incorrectness #
            # bad spacing
            img2 = img.clone()
            img2.set_spacing([6.96]*img.dimension)
            self.assertTrue(not ants.image_physical_space_consistency(img,img2))

            # bad origin
            img2 = img.clone()
            img2.set_origin([6.96]*img.dimension)
            self.assertTrue(not ants.image_physical_space_consistency(img,img2))

            # bad direction
            img2 = img.clone()
            img2.set_direction(img.direction*2)
            self.assertTrue(not ants.image_physical_space_consistency(img,img2))

            # bad dimension
            ndim = img.dimension
            img2 = ants.from_numpy(np.random.randn(*tuple([69]*(ndim+1))).astype('float32'))
            self.assertTrue(not ants.image_physical_space_consistency(img,img2))

            # only one image
            with self.assertRaises(Exception):
                ants.image_physical_space_consistency(img)
            # not an ANTsImage
            with self.assertRaises(Exception):
                ants.image_physical_space_consistency(img, 12)

        # false because of components
        vecimg = ants.from_numpy(np.random.randn(69,70,3), has_components=True)
        vecimg2 = ants.from_numpy(np.random.randn(69,70,4), has_components=True)
        self.assertTrue(not ants.image_physical_space_consistency(vecimg, vecimg2, datatype=True))

    def test_image_type_cast(self):
        # test with list of images
        imgs2 = ants.image_type_cast(self.imgs)
        for img in imgs2:
            self.assertTrue(img.pixeltype, 'float')

        # not a list or tuple
        with self.assertRaises(Exception):
            ants.image_type_cast(self.imgs[0])

        # cast from unsigned int
        imgs = [i.clone() for i in self.imgs]
        imgs[0] = imgs[0].clone('unsigned int')
        imgs2 = ants.image_type_cast(imgs)


    def test_allclose(self):
        for img in self.imgs:
            img2 = img.clone()
            self.assertTrue(ants.allclose(img,img2))
            self.assertTrue(ants.allclose(img*6.9, img2*6.9))
            self.assertTrue(not ants.allclose(img, img2*6.9))

    def test_pickle(self):
        import ants
        import pickle
        img = ants.image_read( ants.get_ants_data("mni"))
        img_pickled = pickle.dumps(img)
        img2 = pickle.loads(img_pickled)

        self.assertTrue(ants.allclose(img, img2))
        self.assertTrue(ants.image_physical_space_consistency(img, img2))

        img = ants.image_read( ants.get_ants_data("r16"))
        img_pickled = pickle.dumps(img)
        img2 = pickle.loads(img_pickled)

        self.assertTrue(ants.allclose(img, img2))
        self.assertTrue(ants.image_physical_space_consistency(img, img2))


if __name__ == '__main__':
    run_tests()
